
#include <map>

#include "Riostream.h"
#include "TString.h"
#include "TFile.h"
#include "TTree.h"

#include "fasterac/fasterac.h"
#include "fasterac/utils.h"
#include "fasterac/farray.h"
#include "fasterac/fast_data.h"
#include "fasterac/qdc.h"
#include "fasterac/rf.h"
#include "fasterac/adc.h"
#include "fasterac/adc_trap.h"
#include "fasterac/jdb_hv.h"
#include "fasterac/scaler.h"

#include "autoroot/F_Data.h"
#include "autoroot/F_Unknown.h"
#include "autoroot/F_Group.h"
#include "autoroot/F_QDC.h"
#include "autoroot/F_QDC2.h"
#include "autoroot/F_QDC3.h"
#include "autoroot/F_QDC4.h"
#include "autoroot/F_QDC_Cnt.h"
#include "autoroot/F_RF.h"
#include "autoroot/F_RF_Cnt.h"
#include "autoroot/F_ADC.h"
#include "autoroot/F_ADC_Trap.h"
#include "autoroot/F_ADC_Cnt.h"
#include "autoroot/F_HV.h"
#include "autoroot/F_Scaler.h"
#include "autoroot/F_Scaler_Cnt.h"


using namespace std;

   F_Data* New_Data (const faster_data_p data)
   {
      F_Data*       out;
      unsigned char alias;
      alias = faster_data_type_alias (data);
      switch (alias) {
         case GROUP_TYPE_ALIAS :
            out = new F_Group ();
            break;
         case QDC_X1_TYPE_ALIAS :
         case QDC_TDC_X1_TYPE_ALIAS :
            out = new F_QDC ();
            break;
         case QDC_X2_TYPE_ALIAS :
         case QDC_TDC_X2_TYPE_ALIAS :
            out = new F_QDC2 ();
            break;
         case QDC_X3_TYPE_ALIAS :
         case QDC_TDC_X3_TYPE_ALIAS :
            out = new F_QDC3 ();
            break;
         case QDC_X4_TYPE_ALIAS :
         case QDC_TDC_X4_TYPE_ALIAS :
            out = new F_QDC4 ();
            break;
         case QDC_COUNTER_TYPE_ALIAS :
            out = new F_QDC_Cnt ();
            break;
         case RF_DATA_TYPE_ALIAS :
            out = new F_RF ();
            break;
         case RF_COUNTER_TYPE_ALIAS :
            out = new F_RF_Cnt ();
            break;
         case ADC_DATA_TYPE_ALIAS :
            out = new F_ADC ();
            break;
         case ADC_TRAP_DATA_TYPE_ALIAS :
            out = new F_ADC_Trap ();
            break;
         case ADC_COUNTER_TYPE_ALIAS :
            out = new F_ADC_Cnt ();
            break;
         case JDB_HV_TYPE_ALIAS :
            out = new F_HV ();
            break;
         case SCALER_MEASUREMENT_TYPE_ALIAS :
            out = new F_Scaler ();
            break;
         case SCALER_COUNTER_TYPE_ALIAS :
            out = new F_Scaler_Cnt ();
            break;
         default :
            out = new F_Unknown ();
            std::cout << "F_Data factory - unknown data type " <<  (int) alias << "." << std::endl;
      }
      out->Initialize (data);
      return out;
   }


   //
   //--  F_Data* indexed by labels  ---------------------------------------------------//
   //
   //            add    - add a new F_Data* (if not already exists) initialized by a faster_data
   //            reset  - reset contents
   //            branch - add a branch to a TTree foreach F_Data*
   //            fill   - adjust contents from a faster_data
   //            free   - delete all F_Data*
   //

   typedef std::map <Int_t, F_Data*> Label_Data_Map;


   void label_data_map_add (Label_Data_Map &ldm, faster_data_p data) {
      faster_buffer_reader_p group_reader;
      void*                  group_buffer;
      faster_data_p          group_data;
      unsigned short         lsize;
      unsigned char          alias = faster_data_type_alias (data);
      unsigned short         label = faster_data_label      (data);
      if (ldm.count (label) == 0) {
         ldm [label] = New_Data (data);
      }
      if (alias == GROUP_TYPE_ALIAS) {
         lsize        = faster_data_load_size     (data);
         group_buffer = faster_data_load_p        (data);
         group_reader = faster_buffer_reader_open (group_buffer, lsize);
         while ((group_data = faster_buffer_reader_next (group_reader)) != NULL) {
            label_data_map_add (ldm, group_data);
         }
         faster_buffer_reader_close (group_reader);
      }
   }


   void label_data_map_display (Label_Data_Map &ldm) {
      cout << "\t LABEL \t DATA NAME" << endl;
      for (Label_Data_Map::iterator ii=ldm.begin(); ii!=ldm.end(); ++ii) {
         cout << " \t " << (*ii).first << " \t " << (*ii).second->GetName () << endl;
      }
   }


   void label_data_map_reset (Label_Data_Map &ldm) {
      for (Label_Data_Map::iterator it = ldm.begin (); it != ldm.end (); ++it) (*it).second->Reset ();  // reset branches
   }


   void label_data_map_branch (Label_Data_Map &ldm, TTree* tree) {
      F_Data* fd;
      for (Label_Data_Map::iterator it = ldm.begin (); it != ldm.end (); ++it) {
        fd = (*it).second;
        tree->Branch (fd->GetName (), fd, 64000, 2);
      }
   }


   void label_data_map_fill (Label_Data_Map &ldm, faster_data_p data, Int_t group_label) {
      faster_buffer_reader_p group_reader;
      void*                  group_buffer;
      faster_data_p          group_data;
      unsigned short         lsize;
      unsigned char          alias = faster_data_type_alias (data);
      unsigned short         label = faster_data_label      (data);
      ldm [label]->Reset (data);
      if (group_label) ldm [label]->Set_Group (group_label);
      if (alias == GROUP_TYPE_ALIAS) {
         lsize        = faster_data_load_size     (data);
         group_buffer = faster_data_load_p        (data);
         group_reader = faster_buffer_reader_open (group_buffer, lsize);
         while ((group_data = faster_buffer_reader_next (group_reader)) != NULL) {
            label_data_map_fill (ldm, group_data, label);
         }
         faster_buffer_reader_close (group_reader);
      }
   }


   void label_data_map_free (Label_Data_Map &ldm) {
     for (Label_Data_Map::iterator it = ldm.begin (); it != ldm.end (); ++it) delete (*it).second;
   }




   //---  Faster to Root Tree automatic converter  --------------------------------------------//


   int main (int argc, char** argv) {

      //!!!!!!!!!!!!!!!
      //cout << "nb_data=" << F_Data::nb_data << endl;
      //!!!!!!!!!!!!!!!

      //  Command line arguments
      if (argc < 3) {
         cout << endl;
         cout << "  usage  :  " << endl;
         cout << "            " << argv [0] << "  input_file.fast   output_file.root" << endl;
         cout << endl;
         return EXIT_FAILURE;
      }
      TString faster_filename = argv [1];
      TString root_filename   = argv [2];

      //  faster input data : memory allocation + read file + data array
      char*    data_space;
      size_t   space_size;
      int      alloc_error;
      farray*  far;
      alloc_error = farray_data_file_to_memory (faster_filename, &data_space, &space_size);
      if (alloc_error == 1) {
         cout  << "error opening " << faster_filename << endl;
         throw 1;
      } else if (alloc_error == 2) {
         cout << "memory error\n";
         throw 2;
      }
      far = farray_new (data_space, space_size);

      //  First data scan -> create a F_Data* for each label
      Label_Data_Map  ldm;          // label and F_Data* map
      for (int i=0; i<far->nb_data; i++) {
         label_data_map_add (ldm, far->data_array [i]);
      }
      //label_data_map_display (ldm);

      //!!!!!!!!!!!!!!!
      //cout << "nb_data=" << F_Data::nb_data << endl;
      //!!!!!!!!!!!!!!!

      //  TTree creation - One branch per F_Data*
      TString title = TString ("AutoRoot from ") + faster_filename;
      TFile *ftree  = new TFile (root_filename, "recreate");
      TTree *tree   = new TTree ("AutoRoot", title);
      label_data_map_branch (ldm, tree);

      //  Second data scan -> fill tree
      Int_t group = 0;
      for (int i=0; i < far->nb_data; i++) {
         label_data_map_reset (ldm);
         label_data_map_fill  (ldm, far->data_array  [i], group);
         tree->Fill           ();
      }

      //  Display result
      cout << endl;
      cout << "   AUTOROOT : " << faster_filename << " => " << root_filename << endl;
      cout << endl;
      tree->Print ();
      cout << endl;

      //  Memory deallocation
      tree->Write         (tree->GetName (), TObject::kWriteDelete);    // save tree
      ftree->Purge        ();
      ftree->Close        ();
      label_data_map_free (ldm);

      //!!!!!!!!!!!!!!!
      //cout << "nb_data=" << F_Data::nb_data << endl;
      //!!!!!!!!!!!!!!!
      farray_free         (far);
      free                (data_space);

   }


